
# Definitions

The [lychee.Definition](/libraries/crux/source/Definition.js)
format allows a cross-platform way to define either a
`Callback`, a `Composite` or a `Module`.

It also integrates [Feature Detection](./feature-detection.md)
mechanisms via the `supports()` callback in order to
allow platform-specific Definitions that integrate
with the matching `platform` tag of the Definition
or Environment.


## Basic Layout

- `lychee.define('identifier')` returns an instance of `lychee.Definition(identifier)` while it exports the Definition to the active [Environment](./environments.md).
- `.tags({ tag: 'value' })` tags a Definition with specific values and integrates with [Feature Detection](./feature-detection.md).
- `.attaches({ name: Asset })` injects an Asset instance as `attachments[name]`.
- `.requires([ 'dependency' ])` requires another Definition via its `identifier`.
- `.includes([ 'dependency' ])` includes another Definition via its `identifier`. This only works with `Composite` Definitions.
- `.supports((lychee, global) => {})` defines the [Feature Detection](./feature-detection.md) and returns `true` or `false`.
- `.exports((lychee, global, attachments) => {})` exports an Implementation and returns either a `Callback`, a `Composite` or a `Module`.

A Definition consists of multiple files. Files that are not
part of the Implementation are mapped as the `attachments`
parameter of the `exports()` callback with their respective
[Asset](./lycheejs-crux.md) instance.

The filesystem structure of a Definition:

```bash
/projects/example            - Project Root Folder
|> /source
|  |> /ui                    - app.ui Namespace
|  |  |> Definition.js       - app.ui.Definition       (Implementation)
|  |  |> Definition.fnt      - attachments['fnt']      (Font instance)
|  |  |> Definition.json     - attachments['json']     (Config instance)
|  |  |> Definition.png      - attachments['png']      (Texture instance)
|  |  |> Definition.name.png - attachments['name.png'] (Texture instance)
```

The name of the file influences the name of the equivalent
`attachments[name]` property. A file named `Definition.example.png`
leads to `attachments['example.png']` being a `Texture`
instance.

## Basic Example

This Example Definition inherits its behaviour from a
`Composite` named `lychee.ui.Layer` and overwrites the
`setCustom()` method, but reuses the `render()` method
from `lychee.ui.Entity`.

In general, all Definitions in the lychee.js Engine
stack use the `Composite Pattern` and not ECMAScript
classes. The Composite pattern allows flexible
combinations of given implementations and the efficient
reintegration of `states` that cause different behaviours.


Example content of `/source/ui/Definition.js`:

```javascript
lychee.define('app.ui.Definition').tags({
	variant: 'awesome'
}).requires([
	'lychee.ui.Entity'
]).includes([
	'lychee.ui.Layer'
]).exports((lychee, global, attachments) => {

	const _Entity = lychee.import('lychee.ui.Entity');
	const _Layer  = lychee.import('lychee.ui.Layer');

	const _CONFIG  = attachments['name.json']; // Config instance
	const _TEXTURE = attachments['png'];       // Texture instance
	const _FONT    = attachments['fnt'];       // Font instance



	/*
	 * IMPLEMENTATION
	 */

	const Composite = function(data) {
	
		let states = Object.assign({}, data);


		this.custom = null; // default value for Object instances

		this.setCustom(states.custom);

		delete states.custom;


		// Composite inherits from _Layer (see includes)
		_Layer.call(this, states);

		states = null;

	};


	Composite.prototype = {

		// XXX: Missing serialize() and deserialize() methods
		// automatically generated by lychee.js Strainer

		render: function(renderer, offsetX, offsetY) {

			// Composite reuses render() method from _Entity (see requires)
			_Entity.prototype.render.call(this, renderer, offsetX, offsetY);

			// Composite introduces custom state
			if (this.custom !== null) {
				this.custom.render(renderer, offsetX, offsetY);
			}

		},

		// Composite introduces custom state and setCustom() method
		setCustom: function(custom) {

			custom = lychee.interfaceof(_Entity, custom) ? custom : null;


			if (custom !== null) {

				this.custom = custom;

				return true;

			}


			return false;

		}

	};


	return Composite;

});
```

